#include "context.h"
#include "config.h"
#include "pool.h"

/*************
Context : Just track seen symbols, counts & escapes; no coding here

        The only fudgy bit here is the Update_See:
        We actually do this for the *2nd* time on
        contexts we code from, since EncodeEscape
        does an additional Adjust for us however,
        the states we update here may not be the
        same ones we used, since the actual coding
        & LOE use excluded states..

*************/

//-----------------------------------------
// static pools
static Pool* ContextPool          = NULL;
static int      ContextPoolCount     = 0;
static Pool* ContextNodePool      = NULL;
static int      ContextNodePoolCount = 0;
//-----------------------------------------

/* */
Context* context_create( Context* parent, uint index ) {

    int      order;
    Context* cntx = autoPool_getHunk( &ContextPool, &ContextPoolCount, sizeof(Context) );   assert(cntx);

    // All is zero:
    assert( cntx->upex.syms == NULL );
    node_null( cntx );
    node_null( &cntx->LRU );

    cntx->parent = parent;
    cntx->index = index;

    if (parent) order = parent->order + 1;
    else        order = 0;

    cntx->order = order;

    if (parent) {
        if (parent->child) {
            node_addHead(parent->child,cntx);
        } else {
            parent->child = cntx;
        }
    }       
    return cntx;
}

/* */
Context * context_find( Context* parent, ulong index ) {
    Context *base = parent->child;
    if (base) {
        Context *c = base;
        for(;;) {
            if (c->index == index) {
                // found it, do MTF
                // this is really easy, cuz we have a circular list,
                //      we just change the start point!
                parent->child = c;
                return c;
            }
            c = node_next(c);
            if (c == base)    break;
        }
    }
    return NULL;
}

/* */
void context_destroyAllContexts( void ) {
    autoPool_destroy( &ContextPool, &ContextPoolCount );
    autoPool_destroy( &ContextNodePool, &ContextNodePoolCount );
}

/* */
void context_destroy( Context* cntx ) {
    ContextNode* syms;
    ContextNode* next;

    if (!cntx)   return;

    node_cut( cntx );
    node_cut( &cntx->LRU );

    syms = cntx->full.syms;
    while (syms) {
        next = syms->next;
        autoPool_freeHunk( &ContextNodePool, &ContextNodePoolCount, syms );
        syms = next;
    }
    syms = cntx->upex.syms;
    while (syms) {
        next = syms->next;
        autoPool_freeHunk( &ContextNodePool, &ContextNodePoolCount, syms );
        syms = next;
    }
        
    autoPool_freeHunk( &ContextPool, &ContextPoolCount, cntx );
}

/* */
void contextData_update( ContextData* cntx, int sym, int order, ulong index, See* see, Context* context ) {
    ContextNode *node;
    bool escape;

    /* <> could track the 'if I had coded' entropy here */
        
    if (cntx->totSymCount >= CONTEXT_CHAR_COUNT_SCALE_DOWN) {
        contextData_halve(cntx);
    }

    cntx->lastSym = sym;

    node = cntx->syms;
    while (node) {
        if (node->sym == sym) {
            if (node->count <= CONTEXT_SYM_INC_NOVEL) {
                cntx->escapeCount -= CONTEXT_ESCP_INC;
                node->count += CONTEXT_SYM_INC - CONTEXT_SYM_INC_NOVEL;
                cntx->totSymCount += CONTEXT_SYM_INC - CONTEXT_SYM_INC_NOVEL;
                if (cntx->escapeCount <= 0) {
                    cntx->escapeCount = 1;
                }
            }
            node->count += CONTEXT_SYM_INC;
            cntx->totSymCount += CONTEXT_SYM_INC;

            escape = FALSE;
        
            goto gotNode;
        }
        node = node->next;
    }

    escape = TRUE;

    node = autoPool_getHunk( &ContextNodePool, &ContextNodePoolCount, sizeof(ContextNode) );
    assert(node);

    node->next = cntx->syms;
    cntx->syms = node;

    node->sym = sym;
    node->count = CONTEXT_SYM_INC_NOVEL;
    cntx->totSymCount += CONTEXT_SYM_INC_NOVEL;       

    if (cntx->escapeCount < CONTEXT_ESCAPE_MAX) {
        cntx->escapeCount += CONTEXT_ESCP_INC;
    }

    ++ cntx->numSyms;

 gotNode: //---------------

    cntx->largestCount = max(cntx->largestCount,node->count);

    if (see) {
        // note that this may or may not be the same state that we coded from, because
        //      of exclusions & such
        see_adjustState( see, cntx->seeState, escape );
        cntx->seeState = see_getState( see, cntx->escapeCount, cntx->totSymCount, index, order, cntx->numSyms, context );
    } else {
        cntx->seeState = NULL;
    }
}

/* */
void context_update( Context* cntx, int sym, ulong index, See* see, int codedOrder ) {

    assert( ! cntx->parent || cntx->parent->order == (cntx->order - 1) );
    assert( cntx->order == PZIP_ORDER || ! cntx->child  || cntx->child->order  == (cntx->order + 1) );
    assert( ((Context *)node_next(cntx))->order == cntx->order );

    /* Fulls don't do See Updates: */
    contextData_update( &cntx->full, sym, cntx->order, index, NULL, cntx );

    if (cntx->order >= codedOrder) {
        contextData_update( &cntx->upex, sym, cntx->order, index, see, cntx );
    }
    /* Not necessarily true cuz of halvings: */
    assert( cntx->full.totSymCount >= cntx->upex.totSymCount || cntx->full.totSymCount >= ((CONTEXT_CHAR_COUNT_SCALE_DOWN>>1)-256) );
    assert( cntx->full.numSyms >= cntx->upex.numSyms || cntx->full.totSymCount >= ((CONTEXT_CHAR_COUNT_SCALE_DOWN>>1)-256));
}

/* */
void contextData_halve( ContextData* cntx ) {
    ContextNode*  node;
    ContextNode** nodePtr;

    cntx->totSymCount = 0;
    cntx->numSyms = 0;
    cntx->largestCount = 0;

    nodePtr = &cntx->syms;
    while( (node = *nodePtr) != NULL ) {
        node->count = (node->count)>>1;

        if (node->count == 0) {
            *nodePtr = node->next;
            autoPool_freeHunk(&ContextNodePool,&ContextNodePoolCount,node);
        } else {
            if (node->count <= CONTEXT_SYM_INC_NOVEL) {
                node->count = CONTEXT_SYM_INC_NOVEL + 1;
            }
            cntx->totSymCount += node->count;

            cntx->numSyms++;

            cntx->largestCount = max(cntx->largestCount,node->count);

            nodePtr = &node->next;
        }
    }
        
    cntx->escapeCount = ((cntx->escapeCount)>>1) + 1;
}

/* */
void contextData_getExcludedInfo( ContextData* cd, Exclude* exc, int* pTotCount, int* pLargestCount, int* pEscapeCount) {
    int largestCount;
    int totCount;
    int escapeCount;

    if (exclude_isEmpty(exc)) {
        totCount = cd->totSymCount;
        largestCount = cd->largestCount;
        escapeCount = cd->escapeCount;
    } else {
        ContextNode * n;

        // escape from un-excluded counts
        //      also count the excluded escape syms, but not as hard
        // rig up the counding so that 1 excluded -> 1 final count

        // helped paper2 2.193 -> 2.188 bpc !!
        // you can get 0.001 bpc by tweaking all these constants :

        // these counts make 1 exc -> 1, 2 -> 2, and 3 -> 2 , etc.
        //      so for low-escape contexts, we get the same counts, and for low-orders we get
        //      much lower escape counts

        largestCount = 0;
        totCount = 0;
        escapeCount = CONTEXT_EXCLUDED_ESCAPE_INIT;

        for (n = cd->syms;   n;   n = n->next) {
            if (!isExcluded( exc, n->sym ) ) {
                totCount += n->count;
                if (n->count > largestCount) {
                    largestCount = n->count;
                }
                if (n->count <= CONTEXT_SYM_INC_NOVEL) {
                    escapeCount += CONTEXT_EXCLUDED_ESCAPE_INC;
                }
            } else {
                if (n->count <= CONTEXT_SYM_INC_NOVEL) {
                    escapeCount += CONTEXT_EXCLUDED_ESCAPE_EXCLUDEDINC;
                }
            }
        }

        escapeCount >>= CONTEXT_EXCLUDED_ESCAPE_SHIFT;
    }               

    *pTotCount = totCount;
    *pLargestCount = largestCount;
    *pEscapeCount = escapeCount;
}

/* */
bool context_chooseFull( Context* c, Exclude* exc, See* see, ulong cntx ) {
    int fLargestCount,fTotCount,fEscapeCount,fRating;
    int uLargestCount,uTotCount,uEscapeCount,uRating;
    SeeState * ss;
        
    return FALSE; // 'text100k' : full helps 28975 to 28967

    // don't use non-det fulls
//      if ( fLargestCount != fTotCount )
    if (c->full.numSyms > 1)        return FALSE;
    if (c->full.totSymCount < 16)   return FALSE;

    contextData_getExcludedInfo( &c->full, exc, &fTotCount, &fLargestCount, &fEscapeCount );
    if (fTotCount == 0) { return FALSE; }

    contextData_getExcludedInfo( &c->upex, exc, &uTotCount, &uLargestCount, &uEscapeCount );
    if (uTotCount == 0) { return TRUE; }

    if ( fTotCount < fEscapeCount ) ss = NULL;
    else                            ss = see_getState( see, fEscapeCount, fTotCount, cntx, c->order, c->full.numSyms, c );

    fRating = ((PZIP_INTPROB_ONE - see_getEscapeP( see, ss, fEscapeCount, fTotCount ))
               * fLargestCount ) / fTotCount;
    
    if ( uTotCount < uEscapeCount ) ss = NULL;
    else                            ss = see_getState( see, uEscapeCount, uTotCount, cntx, c->order, c->upex.numSyms, c );

    uRating = ((PZIP_INTPROB_ONE - see_getEscapeP(see,ss,uEscapeCount,uTotCount))
               * uLargestCount ) / uTotCount;

    // and even some more :
//      fRating >>= 1;

    if (uRating >= fRating) return FALSE;

    return TRUE;
}

