#ifndef CONTEXT_H
#define CONTEXT_H

#include "inc.h"
#include "list.h"
#include "see.h"
#include "exclude.h"

#ifndef DEFINED_CONTEXT
typedef struct Context Context;
#define DEFINED_CONTEXT
#endif

typedef struct ContextNode ContextNode;
typedef struct ContextData ContextData;

struct ContextNode {
    ContextNode* next;
    uword        sym;
    uword        count;
};

/***

sisters, parent & child make this a 3-branch tree :

(n->child->parent == n) ; parent is just redundant

to find the child you want, step onto the child pointer, then
 walk the sisters till you get the ->index you seek

all sisters have the same parent

(semi-hack : the child pointer of the deepest node is actually a Det context!)

***/

struct ContextData {
    int          escapeCount;
    int          numSyms;
    int          totSymCount;
    int          lastSym;
    int          largestCount;
    ContextNode* syms;
    SeeState*    seeState;
};

struct Context {

    Node sisters; /* Must be at head of struct. */

    uint     index;   /* To get to me               */
    int      order;

    Context* parent;
    Context* child;

    ContextData upex;
    ContextData full; /* <> seems pointsless        */

    Node LRU;
};

/*******/

Context* context_create( Context* parent, uint index );
void     context_destroy( Context* c );
void     context_destroyAllContexts( void ); // global & naughty, but oh-so-fast

Context* context_find( Context* parent, ulong index );

void     context_update( Context* c ,int sym, ulong cntx, See* see, int codedOrder);

void     contextData_update( ContextData* cntx, int sym, int order, ulong index, See* see, Context* context );
void     contextData_halve( ContextData* cntx );

void     contextData_getExcludedInfo( ContextData* cd, Exclude* exc, int* pTotCount, int* pLargestCount, int* pEscapeCount );

bool     context_chooseFull( Context* c, Exclude* exc, See* see, ulong cntx );

#endif // CONTEXTS_H
