#include "loe.h"
#include "config.h"
#include "intmath.h"

#define DO_LOE_PENALIZE_NONDET // yep this helps; it's freaky

/*********

 todo : weight between several LOE schemes based on performance

 eg. rating = LOE_success1 * rating1 + ...

**********/

//------------------------------------------------------
// loe_chooseOrder

int loe_chooseOrder( Context** contexts, ulong cntx, int maxOrder, Exclude* exc, See* see, bool useFull ) {
    int choseRating, choseOrder;
    int      rating,      order;

    if (Pzip_LOEType == LOETYPE_NONE)   return maxOrder;

    choseOrder  = 0;
    choseRating = 0;

    for (order = maxOrder;   order >= 0;   --order) {
        Context*     c;
        ContextData* cd;
        int largestCount, totCount, escapeCount;
        SeeState*  ss;

        if (order == 0 && choseRating == 0) {
            return 0; //early out; this is our only choice
        }
        c = contexts[ order ];

        if (!c || c->full.totSymCount == 0) {
            assert( c->upex.totSymCount == 0 );
            continue;
        }

        if (useFull && context_chooseFull(c,exc,see,cntx))   cd = &c->full;
        else                                                 cd = &c->upex;
                
        contextData_getExcludedInfo( cd, exc, &totCount, &largestCount, &escapeCount );
                
        if (totCount == 0)   continue;

        assert( largestCount >= 0 );


        #ifdef DO_LOE_PENALIZE_NONDET
        // ask if its det before exclusion or after?
        //  doesn't seem to make much difference
        if (cd->numSyms > 1) { // this is before
            //if ( largestCount != totCount ) // this is after
            totCount += escapeCount;
        }
        // note : this makes us a use a different seeState for LOE than we do for coding!
        #endif

        if (totCount < escapeCount) ss = NULL;
        else                        ss = see_getState( see, escapeCount, totCount, cntx, order, cd->numSyms, c );

        //cd->loeSeeState = ss; //(disabled, see codecntx.c)

        if (Pzip_LOEType == LOETYPE_MPS) {
            rating = ((PZIP_INTPROB_ONE - see_getEscapeP(see,ss,escapeCount,totCount))
                      * largestCount ) / totCount;
        } else {
            int mpsp;
            int mpsb;

            // A little pseudo-entropy
            // better on some files

            // <> todo : track performance of various LOE schemes & make a weighted rating :
            //      full_rating = Sum[i] weight(i) * rating(i)

            mpsp = ((PZIP_INTPROB_ONE - see_getEscapeP(see,ss,escapeCount,totCount))
                    * largestCount ) / totCount;

            assert( mpsp > 0 && mpsp < PZIP_INTPROB_ONE );

            mpsb = ((PZIP_INTPROB_SHIFT<<4) - ilog2x16(mpsp)); // = 16 * log(1/mpsp)
            rating = mpsp * mpsb + ( PZIP_INTPROB_ONE - mpsp ) * 128;
            rating = INT_MAX - rating;
        }

        if (rating > choseRating) {
            choseRating = rating;
            choseOrder  = order;
        }
    }

    return choseOrder;
}
